#include "circbuf.h"
#include <stdlib.h>

struct CircBuf_t {
    ValueType *array;   /* actual data storage */
    int size;           /* total buffer size */
    int nelems;         /* number of elements currently stored */
    int front;          /* index to front element */
    int back;           /* index to back element */
};

CircBuf *create_buffer(int size)
/* Allocate memory and initialise buffer. */
{
    CircBuf *cb = (CircBuf *)malloc(sizeof(CircBuf));
    if (cb == NULL)
        return NULL;
    
    cb->array = (ValueType *)malloc(size * sizeof(ValueType));
    if (cb->array == NULL) {
        free(cb);
        return NULL;
    }

    cb->size = size;
    cb->nelems = 0;
    cb->front = 0;
    cb->back = 0;
    return cb;
}

void destroy_buffer(CircBuf *cb)
/* Deallocate memory. */
{
    free(cb->array);
    free(cb);
}

int add_item(CircBuf *cb, const ValueType *item)
/* Add a new element to front of buffer.
 * Returns 0 for success, and -1 if buffer is full. */
{
    if (cb->nelems == cb->size)
        return -1;

    cb->array[cb->front] = *item;
    if (++cb->front == cb->size) /* wrap around */
        cb->front = 0;
    ++cb->nelems;
    return 0;
}

#if 0
// alternative version: uses modulus operator
int add_item(CircBuf *cb, const ValueType *item)
/* Add a new element to front of buffer.
 * Returns 0 for success, and -1 if buffer is full. */
{
    if (cb->nelems == cb->size)
        return -1;

    cb->array[cb->front] = *item;
    cb->front = (cb->front + 1) % cb->size; /* wrap around */
    ++cb->nelems;
}
#endif

int get_item(CircBuf *cb, ValueType *item)
/* Remove element from back of buffer, and assign it to *item. 
 * Returns 0 for success, and -1 if buffer is empty. */
{
    if (cb->nelems == 0)
        return -1;

    --cb->nelems;
    *item = cb->array[cb->back];
    if (++cb->back == cb->size) /* wrap around */
        cb->back = 0;
    return 0;
}

/* Return number of elements stored in buffer. */
int get_nitems(const CircBuf *cb) { return cb->nelems; }

/* Return total size of buffer. */
int get_size(const CircBuf *cb) { return cb->size; }
